from gtagora.exception import AgoraException
from gtagora.models.base import LinkToFolderMixin, BaseModel, DownloadDatasetMixin, SearchMixin, TagMixin
from gtagora.models.dataset import Dataset
from gtagora.models.timeline import TimelineItem


class Series(LinkToFolderMixin, DownloadDatasetMixin, TagMixin, BaseModel, SearchMixin):
    BASE_URL = '/api/v1/serie/'
    BASE_URL_V2 = '/api/v2/series/'

    def get_datasets(self, filters=None):
        if filters and not isinstance(filters, dict):
            raise AgoraException('The filter must be a dict')

        url = f'{self.BASE_URL}{self.id}/datasets/?limit=10000000000'
        return self._get_object_list(url, filters, Dataset)

    def upload(self, input_files, target_files=None):
        if target_files and len(input_files) != len(target_files):
            raise AgoraException("The Inputfiles and TargetFiles must have the same length")

        if isinstance(input_files, str):
            files = [input_files]
        else:
            files = input_files

        datasets = []
        for index, curFile in enumerate(files):
            cur_target_file = None
            if target_files:
                cur_target_file = target_files[index]
            datasets.append(Dataset.upload_files(self.http_client, curFile, cur_target_file, series_id=self.id))
        return datasets

    def upload_dataset(self, input_files, dataset_type, target_files=None):
        # This function creates a dataset of a given type all files given as input will be added to one dataset.
        # Please note: At the moment there is no consistency check. We could create datasets with improper
        # files (e.g. a PAR/REC dataset without PAR/REC files)
        return self.http_client.upload_dataset(input_files, target_files, serie_id=self.id, dataset_type=dataset_type)

    def get_parameter(self, name):
        datasets = self.get_datasets()
        for dataset in datasets:
            par = dataset.get_parameter(name)
            if par:
                return par

    def search_parameter(self, search_string):
        datasets = self.get_datasets()
        pars = []
        for dataset in datasets:
            cur_pars = dataset.search_parameter(search_string)
            pars.extend(cur_pars)
        return pars

    def copy_to_folder(self, target_folder_id):
        url = f'{self.BASE_URL_V2}{self.id}/copy_to_project/{target_folder_id}/'
        response = self.http_client.post(url, json={}, timeout=60)
        if response.status_code != 200:
            raise AgoraException(f'Could not copy the series: status = {response.status_code}')

        return self._get_new_series_from_timeline(response)

    def __str__(self):
        return f"Series: {self.name}"

    def _get_new_series_from_timeline(self, response):
        timeline_item = TimelineItem.from_response(response.json(), self.http_client)
        timeline_item = timeline_item.poll()
        if not timeline_item.data:
            return None

        related_objects = timeline_item.data.get('related_objects')
        if related_objects:
            for obj in related_objects:
                if obj.get('id') and obj.get('id') != self.id:
                    try:
                        return self.get(obj.get('id'), self.http_client)
                    except:
                        pass

        return None



import os
import zipfile
import hashlib
from itertools import zip_longest
from pathlib import Path
from urllib.parse import urlparse


def get_file_info(path):
    total_size = 0
    nof_files = 0

    if isinstance(path, str):
        path = Path(path)

    if path.is_file():
        return 1, path.stat().st_size

    for dirpath, dirnames, filenames in os.walk(str(path)):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
        nof_files += len(filenames)
    return nof_files, total_size


class ZipUploadFiles:

    MAX_FILE_LIMIT = 100*1024*1024
    MAX_ZIP_FILE_SIZE = 2*1024*1024*1024

    def __init__(self, input_files, target_files):
        self.input_files = input_files
        self.target_files = target_files if target_files else []
        self._zip_is_required = False

    def create_zip(self, path: Path):
        files_to_zip = self._create_file_list()

        if self._zip_is_required is False:
            return self.input_files, self.target_files

        index = 0
        input_files = []
        target_files = []

        while index < len(files_to_zip):

            zip_filename = f'upload_{index}.agora_upload'
            zip_path = path / zip_filename
            input_files.append(zip_path)
            target_files.append(zip_filename)

            with zipfile.ZipFile(zip_path, 'w', compression=zipfile.ZIP_STORED) as z:
                for file, target_file, do_zip in files_to_zip[index:]:
                    if do_zip:
                        z.write(file, target_file)
                    else:
                        input_files.append(file)
                        target_files.append(target_file)
                    index += 1

                    compressed_size = sum([info.compress_size for info in z.infolist()])
                    if compressed_size > self.MAX_ZIP_FILE_SIZE:
                        break

        return input_files, target_files

    def _create_file_list(self):

        def create_entry(file, target_file):
            size = os.path.getsize(file)
            do_zip = size < self.MAX_FILE_LIMIT

            if do_zip:
                self._zip_is_required = True

            return file, target_file, do_zip

        file_list = [create_entry(file, target_file)
                     for file, target_file in zip_longest(self.input_files, self.target_files)]
        return file_list


def remove_illegal_chars(path: str):
    illegal = [':', '*', '"', '<', '>', '|']

    for i in illegal:
        path = path.replace(i, '')

    return path


def sha1(path: Path):

    sha1 = hashlib.sha1()
    BUF_SIZE = 1024*1024

    with path.open('rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)

    return sha1.hexdigest()

def validate_url(url):
    # check if the url has a scheme. If not then add it
    u = urlparse(url)
    if u.scheme != 'http' and u.scheme != 'https':
        url = 'http://' + url
        u = urlparse(url)

    if u.path:
        url = u.scheme + '://' + u.netloc

    return url



import os

from gtagora.exception import AgoraException
from gtagora.models.base import BaseModel, LinkToFolderMixin, TagMixin
from gtagora.models.datafile import Datafile
from gtagora.models.image_info import ImageInfo
from gtagora.models.parameter import Parameter


class DatasetType:
    NONE = 0
    PHILIPS_RAW = 100
    PHILIPS_PARREC = 101
    PHILIPS_SPECTRO = 102

    BRUKER_RAW = 200
    BRUKER_SUBJECT = 201
    BRUKER_IMAGE = 202
    DICOM = 300
    SIEMENS_RAW = 400
    ISMRMRD = 500
    NIFTI1 = 600
    NIFTI2 = 601
    NIFTI_ANALYZE75 = 602
    JMRUI_SPECTRO = 700
    QUERY = 10000
    OTHER = 100000


class Dataset(LinkToFolderMixin, TagMixin, BaseModel):
    BASE_URL = '/api/v1/dataset/'
    BASE_URL_V2 = '/api/v2/dataset/'

    def _set_values(self, model_dict):
        for key, value in model_dict.items():
            if key == 'datafiles':
                datafiles = []
                for item in model_dict['datafiles']:
                    datafiles.append(Datafile.from_response(item,
                                                            http_client=self.http_client))
                setattr(self, key, datafiles)
            else:
                setattr(self, key, value)

    def get_datafiles(self):
        if hasattr(self, 'datafiles'):
            return self.datafiles
        else:
            return []

    def download(self, filename=None):
        datafiles = self.get_datafiles()
        downloaded_files = []
        for datafile in datafiles:
            downloaded_files.append(datafile.download(filename))
        return downloaded_files

    def get_parameter(self, name):
        url = f'{self.BASE_URL}{self.id}/parameter/?description__name={name}&limit=10000000000'
        pars = self._get_object_list(url, None, Parameter)
        return pars[0] if pars else None

    def get_info(self):
        url = f'{self.BASE_URL}{self.id}/image_info/'
        info = self._get_object_list(url, None, ImageInfo)
        return info if info else None

    def search_parameter(self, search_string):
        url = f'{self.BASE_URL}{self.id}/parameter/?search={search_string}&limit=10000000000'
        return self._get_object_list(url, None, Parameter)

    def create(self, series_id=None, exam_id=None, folder_id=None, type=DatasetType.OTHER):
        url = Dataset.BASE_URL
        if series_id:
            data = {"serie": series_id, "type": type}
        elif exam_id:
            data = {"exam": exam_id, "type": type}
        elif folder_id:
            data = {"folder": folder_id, "type": type}
        else:
            raise AgoraException('Please specify a SeriesID, ExamID or FolderID')

        response = self.http_client.post(url, json=data, timeout=60)
        if response.status_code == 201:
            data = response.json()
            if 'id' in data:
                self._set_values(data)
                return self

        raise AgoraException('Could not create the dataset')

    @classmethod
    def upload_files(cls, http_client, input_files, target_files, series_id=None, exam_id=None, folder_id=None,
                       dataset_type=DatasetType.OTHER):
        if (not series_id and not folder_id and not exam_id):
            raise AgoraException('Please specify a SeriesID, ExamID or FolderID')

        if isinstance(input_files, str):
            Files = []
            Files.append(input_files)
        else:
            Files = input_files

        if not target_files:
            target_files = [os.path.basename(file) for file in Files]

        if len(target_files) < len(Files):
            raise AgoraException('TargetFiles list too short.')

        for curFile in Files:
            if not os.path.isfile(curFile):
                raise AgoraException('File does not exist: ' + curFile)

        dataset = Dataset(http_client=http_client)
        dataset.create(series_id=series_id, exam_id=exam_id, folder_id=folder_id, type=dataset_type)
        url = f'/api/v1/dataset/{dataset.id}/upload/'
        if http_client.upload(url, input_files, target_files):
            return Dataset.get(dataset.id, http_client)

        raise AgoraException("Failed to create a new dataset")

    def copy_to_folder(self, target_folder_id):
        url = f'{self.BASE_URL_V2}{self.id}/copy_to_project/{target_folder_id}/'
        response = self.http_client.post(url, json={}, timeout=60)
        if response.status_code != 200:
            raise AgoraException(f'Could not copy the dataset: status = {response.status_code}, message = {response.text}')
        return Dataset.from_response(response.json(), self.http_client)

    def __str__(self):
        return f"Dataset: {self.name}, {self.total_size}"



from requests.auth import HTTPBasicAuth

from gtagora.http.auth import ApiKeyAuth, TokenAuth, NoAuth


class Connection:
    def __init__(self, url, verify_certificate=True):
        self.url = url
        self.verify_certificate = verify_certificate

    def get_auth(self):
        raise NotImplementedError


class BasicConnection(Connection):
    def __init__(self, url, user, password, verify_certificate=True):
        super().__init__(url, verify_certificate=verify_certificate)
        self.user = user
        self.password = password

    def get_auth(self):
        return HTTPBasicAuth(self.user, self.password)


class ApiKeyConnection(Connection):
    def __init__(self, url, api_key, verify_certificate=True):
        super().__init__(url, verify_certificate=verify_certificate)
        self.api_key = api_key

    def get_auth(self):
        return ApiKeyAuth(self.api_key)


class TokenConnection(Connection):
    def __init__(self, url, verify_certificate=True):
        super().__init__(url, verify_certificate=verify_certificate)
        self.token = None

    def get_auth(self):
        if self.token:
            return TokenAuth(self.token)
        return NoAuth()

    def login(self, client, user, password):
        response = client.post("/api/v1/rest-auth/login/", data={'username': user, 'password': password})
        if response.status_code == 200:
            self.token = response.json()['key']
        else:
            raise Exception(response.text)


from gtagora.models.base import BaseModel
from gtagora.utils import sha1
from gtagora.exception import DownloadError

import logging
from pathlib import Path

logger = logging.getLogger("gtAgora")


class Datafile(BaseModel):
    BASE_URL = '/api/v1/datafile/'

    def download(self, path: Path):
        # if not filename:
        #     head, tail = os.path.split(self.rel_filename)
        #     filename = tail

        # if os.path.isdir(filename):
        #     filename = os.path.join(filename, self.rel_filename)

        final_path = path / self.original_filename
        final_path.parent.mkdir(parents=True, exist_ok=True)

        if not self.check_for_existing_file(final_path):
            url = f'{self.BASE_URL}{self.id}/download/'
            self.http_client.download(url, final_path.as_posix())

        # downloaded_file = deepcopy(self)
        # downloaded_file.download_path = filename
        return final_path

    def check_for_existing_file(self, desired_path: Path):
        if desired_path.exists():
            if desired_path.is_file():
                if self.size == desired_path.stat().st_size and sha1(desired_path) == self.sha1:
                    return True
            else:
                raise DownloadError(f"File already exists but it's not a file. {desired_path}")
        return False

    def __str__(self):
        return f'{self.original_filename} {self.size}'



from gtagora.exception import AgoraException
from gtagora.models.base import BaseModel
from gtagora.models.dataset import Dataset
from gtagora.models.exam import Exam
from gtagora.models.series import Series


class FolderItem(BaseModel):
    BASE_URL = '/api/v1/folderitem/'

    def _set_values(self, model_dict):
        from gtagora.models.folder import Folder

        for key, value in model_dict.items():
            if key == 'content_object':
                content_object = model_dict['content_object']
            else:
                setattr(self, key, value)

        if self.content_type == 'folder':
            self.object = Folder.from_response(content_object, http_client=self.http_client)
        elif self.content_type == 'exam':
            self.object = Exam.from_response(content_object, http_client=self.http_client)
        elif self.content_type == 'serie':
            self.object = Series.from_response(content_object, http_client=self.http_client)
        elif self.content_type == 'dataset':
            self.object = Dataset.from_response(content_object, http_client=self.http_client)

    def __str__(self):
        return f"FolderItem {self.id}, {self.content_type}, {self.object.name}"



from gtagora.models.base import BaseModel


class Trash(BaseModel):

    BASE_URL = '/api/v1/trash/'

    def get_items(self, project_id):
        items = self.get_list(http_client=self.http_client)
        return [i for i in items if i.project == project_id]

    def empty(self, project_id):
        items = self.get_items(project_id)
        for item in items:
            url = f'{self.BASE_URL}{item.id}/delete'
            response = self.http_client.delete(url, timeout=60)
            if response.status_code != 204:
                print(f'Could not delete item with id = {item.id}')






import datetime
import time

from gtagora.exception import AgoraException
from gtagora.models.base import BaseModel


class TimelineItem(BaseModel):
    BASE_URL = '/api/v2/timeline/'
    TIMEOUT = 60

    def join(self):
        return self.poll()

    def poll(self, interval=2):
        start_time = datetime.datetime.now()
        while (datetime.datetime.now() - start_time).seconds < self.TIMEOUT:
            timeline = self.get(self.id, http_client=self.http_client)

            state = timeline.data.get('state')
            if state:
                if state == 0 or state == 1:
                    time.sleep(interval)
                    continue
                elif state == 2:
                    return timeline
                elif state == 3:
                    raise AgoraException(timeline.error)
                elif state == 4 or state == 5:
                    return None
            else:
                return None




import datetime
import os
import time
import tempfile
from pathlib import Path
from typing import List

from gtagora.models.base import BaseModel
from gtagora.exception import AgoraException
from gtagora.utils import ZipUploadFiles


class ImportPackage(BaseModel):
    mTaskInfoTimeout = 120
    BASE_URL = '/api/v1/import/'

    def __init__(self, http_client=None):
        super().__init__(http_client=http_client)
        self.last_progress = None
        self.zip_upload = False

    def create(self):
        url = self.BASE_URL
        response = self.http_client.post(url, json={}, timeout=60)
        if response.status_code == 201:
            data = response.json()
            if 'id' in data:
                self._set_values(data)
                return self
        raise AgoraException("Can't create an Import object")

    def upload(self, input_files: List[Path], target_folder_id: int = None, json_import_file=None, wait=True,
               timeout: int = None, progress=False, relations: dict = None):

        base_url = '/api/v1/import/' + str(self.id) + '/'
        url = base_url + 'upload/'

        input_files, target_files = self._prepare_paths_to_upload(input_files)
        relations = self._prepare_relations(relations, input_files, target_files)

        if self._check_zip_option(input_files):
            print("Prepare optimized upload")
            with tempfile.TemporaryDirectory() as temp_dir:
                zip_upload = ZipUploadFiles(input_files, target_files)
                input_files, target_files = zip_upload.create_zip(Path(temp_dir))
                response = self.http_client.upload(url, input_files, target_files, progress=progress)
        else:
            self.zip_upload = False
            response = self.http_client.upload(url, input_files, target_files, progress=progress)

        if self.complete(json_import_file, target_folder_id=target_folder_id, relations=relations):
            if wait:
                start_time = datetime.datetime.now()
                while (datetime.datetime.now() - start_time).seconds < timeout if timeout else True:
                    data = self.progress()
                    if data['state'] == 5 or data['state'] == -1:
                        return data
                    time.sleep(5)

        raise AgoraException(f'Failed to complete upload {self.id}: {response.status_code}')

    def complete(self, json_import_file=None, target_folder_id=None, relations: dict = None):
        url = self.BASE_URL + str(self.id) + '/complete/'
        post_data = {}
        if json_import_file:
            post_data.update({'import_file': json_import_file})
        if target_folder_id:
            post_data.update({'folder': target_folder_id})
        if relations:
            post_data.update({'relations': relations})

        response = self.http_client.post(url, json=post_data)
        if response.status_code == 204:
            return True
        raise AgoraException(f'fail to get progress from ImportPackage {self.id}: {response.status_code}')

    def progress(self):
        url = self.BASE_URL + str(self.id) + '/progress/'

        response = self.http_client.get(url)
        if response.status_code == 200:
            data = response.json()
            if 'state' not in data:
                raise AgoraException(f'Could not get the task info: {data}')
            self.last_progress = data
            return self.last_progress
        raise AgoraException(f'fail to get progress from ImportPackage {self.id}: {response.status_code}')

    def get_objects(self):
        url = self.BASE_URL + str(self.id) + '/get_objects/'

        response = self.http_client.get(url)
        if response.status_code == 200:
            data = response.json()
            return data
        raise AgoraException(f'fail to get object from ImportPackage {self.id}: {response.status_code}')

    def _prepare_paths_to_upload(self, paths: List[Path], target_folder_id=None, wait=True, progress=False):
        input_files = []
        target_files = []

        for path in paths:
            if path.is_dir():
                for root, dirs, files in os.walk(path):
                    for f in files:
                        absolute_file_path = Path(root, f)
                        input_files.append(absolute_file_path)
                        target_files.append(absolute_file_path.relative_to(path.parent).as_posix())
            elif path.is_file():
                input_files.append(path.absolute())
                target_files.append(path.name)
            else:
                raise TypeError(f"Can't upload other than a file or a directory: {path.as_posix()}")

        return input_files, target_files

    def _prepare_relations(self, relations, input_files, target_files):
        if not relations:
            return None

        new_relations = dict()
        all_related_files = []
        for relation_target, related_files in relations.items():
            all_related_files.extend(related_files)
        if len(all_related_files) > len(set(all_related_files)):
            raise AgoraException(f'Related files can only occur once')

        for relation_target, related_files in relations.items():
            target_path = self._to_target_path(relation_target, input_files, target_files)
            if not target_path:
                raise AgoraException(f'The relation {relation_target} is not a file which is uploaded')


            new_related_files = []
            for related_file in related_files:
                new_file = self._to_target_path(related_file, input_files, target_files)
                if not new_file:
                    raise AgoraException(f'The related file {related_file} is not a file which is uploaded')
                new_related_files.append(new_file)
            new_relations[target_path] = new_related_files

        return new_relations

    def _check_zip_option(self, input_files):
        return len(input_files) > 5

    def _to_target_path(self, path: str, input_files, target_files):
        try:
            absolute_path = Path(path).absolute()
            if absolute_path in input_files:
                index = input_files.index(absolute_path)
                return target_files[index]
            elif path in target_files:
                return path
            else:
               return None
        except:
            return None


def import_data(http_client, paths: List[Path], target_folder_id: int = None, json_import_file: Path = None, wait=True,
                progress=False, relations: dict =None):
    """
    Import a directory or a list of files with optional target file names.

    The target folder is optional. If
    target_folder is None and data is uploaded that can't be trated as an exam or series a new folder in the
    root will be created.

    :param files: One directory or multiple files as string or Path
    :param target_folder: The target folder
    :param wait: Wait until the upload and import process ha sbeen finished
    :returns: The import package. Can be used to watch the upload
    """
    from gtagora.models.import_package import ImportPackage

    import_package = ImportPackage(http_client=http_client).create()
    print("ImportPackage ID={}".format(import_package.id))

    if json_import_file:
        if not json_import_file.exists():
            raise FileNotFoundError(f"json_import_file {json_import_file} not found")

    import_package.upload(paths,
                          target_folder_id=target_folder_id,
                          json_import_file=json_import_file.name if json_import_file else None,
                          wait=wait,
                          progress=progress,
                          relations=relations)
    return import_package



from gtagora.exception import AgoraException
from gtagora.models.base import BaseModel


class Group(BaseModel):
    BASE_URL = '/api/v1/groups/'

    # def create(self, username, email=None, first_name=None, last_name=None):
    #     url = self.BASE_URL
        
    #     data = {
    #         'username': username,
    #         'email': email,
    #         'first_name': first_name,
    #         'last_name': last_name
    #     }

    #     response = self.http_client.post(url, data, timeout=60)
    #     if response.status_code == 201:
    #         data = response.json()
    #         if 'id' in data:
    #             self._set_values(data)
    #             return self

    #     raise AgoraException('Could not create the user')

    def delete(self):
        raise AgoraException("Can't delete group via the python interface")

    def __str__(self):
        return f"Group {self.id}: {self.name}"



from gtagora.exception import AgoraException
from gtagora.models.base import BaseModel


class Host(BaseModel):
    def copy_to_project(self, project_id):
        url = f'/api/v2/project/{project_id}/host/'
        data = self.to_dict()
        response = self.http_client.post(url, json=data, timeout=60)

        if response.status_code != 201:
            raise AgoraException(f'Cannot copy the task: status = {response.status_code}')

    def move_to_project(self, project_id):
        url = f'/api/v2/project/{project_id}/host/{self.id}/move/'
        response = self.http_client.post(url, json={}, timeout=60)

        if response.status_code != 200:
            raise AgoraException(f'Cannot copy the task: status = {response.status_code}')


from gtagora.exception import AgoraException
from gtagora.models.base import LinkToFolderMixin, BaseModel, TagMixin
from gtagora.models.breadcrumb import Breadcrumb
from gtagora.models.dataset import Dataset
from gtagora.models.datafile import Datafile
from gtagora.models.exam import Exam
from gtagora.models.folder_item import FolderItem
from gtagora.models.import_package import import_data
from gtagora.models.series import Series
from gtagora.utils import remove_illegal_chars

from pathlib import Path
from typing import List
from functools import partial


class Folder(LinkToFolderMixin, TagMixin, BaseModel):
    BASE_URL = '/api/v1/folder/'

    def get_items(self):
        items = []

        url = self.BASE_URL + str(self.id) + '/items/?limit=10000000000'
        response = self.http_client.get(url)

        for item in response.json():
            if 'content_object' in item and 'content_type' in item:
                items.append(FolderItem.from_response(item, http_client=self.http_client))

        return items

    def is_folder(self, name):
        items = self.get_items()
        for item in items:
            if isinstance(item.object, Folder) and item.object.name == name:
                return True

        return False

    def is_subfolder_of(self, folder):
        if isinstance(folder, Folder):
            folder_id = folder.id
        elif isinstance(folder, int):
            folder_id = folder
        else:
            raise AgoraException('The folder argument must either be a Folder class or a folder ID')

        breadcrumb = self.get_breadcrumb()
        for b in breadcrumb:
            if b.object_id == folder_id:
                return True

        return False

    def get_folder(self, path):
        if isinstance(path, str):
            path = Path(path)

        breadcrumb = None
        cur_breadcrumb_index = None
        cur_folder = self
        for part in path.parts:
            if part:
                if part == '..':
                    if not breadcrumb:
                        breadcrumb = cur_folder.get_breadcrumb()
                        cur_breadcrumb_index = len(breadcrumb)-1
                    cur_breadcrumb_index -= 1
                    cur_breadcrumb_index = max(cur_breadcrumb_index, 0)
                elif part == '.':
                    continue
                else:
                    if cur_breadcrumb_index:
                        cur_folder = self._get_object(breadcrumb[cur_breadcrumb_index].object_id)
                        breadcrumb = None
                        cur_breadcrumb_index = None
                    cur_folder = cur_folder._get_by_name(part, Folder)
                    if not cur_folder:
                        return cur_folder

        if cur_breadcrumb_index:
            cur_folder = self._get_object(breadcrumb[cur_breadcrumb_index].object_id)
        return cur_folder

    def get_folders(self, recursive=False):
        folders = []
        items = self.get_items()
        for item in items:
            if isinstance(item.object, Folder):
                folders.append(item.object)
                if recursive:
                    folders = folders + item.get_folders(recursive)

        return folders

    def get_exams(self, recursive=False):
        exams = []
        items = self.get_items()
        for item in items:
            if isinstance(item.object, Exam):
                exams.append(item.object)
            if recursive and isinstance(item, Folder):
                exams = exams + item.get_exams(recursive)

        return exams

    def get_exam(self, name):
        return self._get_by_name(name, Exam)

    def get_series(self, recursive=False):
        series = []
        items = self.get_items()
        for item in items:
            if isinstance(item.object, Series):
                series.append(item.object)
            if recursive and isinstance(item, Folder):
                series = series + item.get_series(recursive)

        return series

    def get_serie(self, name):
        return self._get_by_name(name, Series)

    def get_datasets(self, recursive=False):
        datasets = []
        items = self.get_items()
        for item in items:
            if isinstance(item.object, Dataset):
                datasets.append(item.object)
            if recursive and isinstance(item, Folder):
                datasets.extend(item.get_series(recursive))

        return datasets

    def get_dataset(self, name):
        return self._get_by_name(name, Dataset)


    def get_breadcrumb(self):
        url = f'{self.BASE_URL}{self.id}/breadcrumb/?limit=10000000000'
        return self._get_object_list(url, None, Breadcrumb)

    def path(self):
        breadcrumb = self.get_breadcrumb()
        first = True
        p = ''
        for b in breadcrumb:
            if not first:
                p += '/'
            p+= b.name
            first = False
        return p

    def download(self, target_path: Path, recursive=False) -> List[Datafile]:

        # Get all Exams in the current folder and download them
        downloaded_files = self.download_exams(target_path, recursive=recursive)
        downloaded_files.extend(self.download_series(target_path, recursive=recursive))
        downloaded_files.extend(self.download_datasets(target_path, recursive=recursive))

        return downloaded_files

    def download_exams(self, target_path: Path, recursive=False):
        exams = self.get_exams()
        return self._download_objects(exams, target_path, Folder.download_exams, recursive)

    def download_series(self, target_path: Path, recursive=False):
        exams = self.get_series()
        return self._download_objects(exams, target_path, Folder.download_series, recursive)

    def download_datasets(self, target_path: Path, recursive=False):
        exams = self.get_datasets()
        return self._download_objects(exams, target_path, Folder.download_datasets, recursive)

    def _download_objects(self, objects, target_path: Path, download_fct, recursive=False):
        downloaded_files = [obj.download(target_path) for obj in objects]

        # Download all exams in subfolders as well when the recursive option is true
        if recursive:
            for folder in self.get_folders():
                downloaded_files.extend(partial(download_fct, folder)(
                    target_path / remove_illegal_chars(folder.name), recursive))

        return downloaded_files

    def upload(self, paths: List[Path], wait=True, progress=False, relations: dict =None):
        for path in paths:
            if not path.exists():
                raise FileNotFoundError(path.as_posix())
        return import_data(self.http_client, paths=paths, target_folder_id=self.id, wait=wait, progress=progress, relations=relations)

    def create_folder(self, name):
        url = f'{self.BASE_URL}{self.id}/new/'
        post_data = {"name": name}
        response = self.http_client.post(url, json=post_data)
        if response.status_code == 201:
            data = response.json()
            if 'content_object' in data:
                return Folder.from_response(data['content_object'], http_client=self.http_client)

        raise AgoraException(f'Could not create the folder {name}')

    def get_or_create(self, path):
        if isinstance(path, str):
            path = Path(path)

        next_folder = self
        for part in path.parts:
            next_folder_exists = False
            for folder in next_folder.get_folders():
                if folder.name == part:
                    next_folder = folder
                    next_folder_exists = True
                    break
            if not next_folder_exists:
                next_folder = next_folder.create_folder(part)

        return next_folder

    def delete_item(self, ids):
        for id in ids:
            url = f'/api/v1/folderitem/delete_ids/'
            response = self.http_client.post(url, json={"ids": ids}, timeout=60)

    def parent(self):
        breadcrumb = self.get_breadcrumb()
        if len(breadcrumb) > 1:
            id = breadcrumb[-2].object_id
            return self._get_object(id)
        else:
            return None

    def _get_by_name(self, name, instance):
        items = self.get_items()
        for item in items:
            if isinstance(item.object, instance) and item.object.name == name:
                return item.object

        return None

    def __str__(self):
        return f"Folder: {self.name}"



from pathlib import Path
from typing import List

import urllib3

from gtagora.exception import AgoraException
from gtagora.http.client import Client
from gtagora.http.connection import ApiKeyConnection, TokenConnection
from gtagora.models.dataset import Dataset
from gtagora.models.exam import Exam
from gtagora.models.folder import Folder
from gtagora.models.group import Group
from gtagora.models.import_package import import_data
from gtagora.models.patient import Patient
from gtagora.models.project import Project
from gtagora.models.series import Series
from gtagora.models.tag import Tag
from gtagora.models.task import Task
from gtagora.models.user import User
from gtagora.models.version import Version
from gtagora.utils import validate_url

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class Agora:
    long_timeout = 180
    verify_certificate = False

    default_client = None
    version: Version = None

    def __init__(self, client):
        self.http_client = client
        self.set_default_client(client)
        self.version = self.get_version()
        self.version.needs('6.0.0', error_message='The python interface needs Agora version 6.0.0 or higher. Please update Agora')
        self.import_directroy = self.import_directory # for backward-compatibility.

    @staticmethod
    def create(url, api_key=None, user=None, password=None, token=None):
        """Creates an Agora instance. Prefer this method over using the Agora constructor.

        To authenticate use either the api_key parameter or the user and password parameter.
        Avoid writing your password into python files! Use the API key instead since you can
        simply renew or disable it.

        Arguments:
            url {string} -- The base url of the Agora server (e.g "https://agora.mycompany.com")

        Keyword Arguments:
            api_key {string} -- The API key of  (default: {None})
            user {string} -- The username (default: {None})
            password {string} -- The password (default: {None})

        Returns:
            Agora -- The agora instance
        """
        url = validate_url(url)

        if api_key:
            connection = ApiKeyConnection(url, api_key=api_key, verify_certificate=Agora.verify_certificate)
            client = Client(connection=connection)
        elif token:
            connection = TokenConnection(url, verify_certificate=Agora.verify_certificate)
            client = Client(connection=connection)
            connection.token = token
        else:
            connection = TokenConnection(url, verify_certificate=Agora.verify_certificate)
            client = Client(connection=connection)
            connection.login(client, user, password)

        if not client.check_connection():
            raise AgoraException('Could not connect to the Agora server at ' + url)

        return Agora(client)

    @staticmethod
    def set_default_client(client):
        Agora.default_client = client

    def ping(self):
        """Pings Agora and tests the connection

        Returns:
            bool -- True if the connection was successful.
        """
        url = '/api/v1/version/'
        response = self.http_client.get(url)
        if response.status_code == 200:
            data = response.json()
            return 'server' in data

        return False

    # Project
    def create_project(self, name: str, description: str= '', copy_from_id: int =None):
        """Creates a new project. If "copy_from_id" is specified then all the settings are
           copied from the project with the given id

        Returns:
            Project -- the new project
        """
        data = dict()
        data['name'] = name
        data['description'] = description
        response = self.http_client.post(Project.BASE_URL, json=data, timeout=60)
        if response.status_code == 201:
            data = response.json()
            project = Project.from_response(data, http_client=self.http_client)

            if copy_from_id:
                try:
                    other_project = self.get_project(copy_from_id)
                    project.copy_settings_from(other_project, copy_members=True, copy_tasks=True, copy_hosts=True)
                except:
                    print(f'WARNING: Cannot copy the settings from the project with id={copy_from_id}')

            return project
        else:
            raise AgoraException(f'Could not create the project {name}: status = {response.status_code}')

    def get_or_create_project(self, name: str, description: str= '', copy_from_id: int =None):
        filters = dict()
        filters['name'] = name
        projects = self.get_projects(filters=filters)
        if projects:
            return projects[0]
        return self.create_project(name, description, copy_from_id)


    def get_projects(self, filters=None):
        return Project.get_list(filters, http_client=self.http_client)

    def get_project(self, project_id):
        return Project.get(project_id, http_client=self.http_client)

    def get_myagora(self):
        return Project.get('myagora', http_client=self.http_client)

    # Folder
    def get_root_folder(self):
        """Returns a list of all root folder of the current user.

        Returns:
            List[Folder] -- The list of root folders
        """
        return self.get_myagora().get_root_folder()

    def get_folder(self, folder_id: int):
        """Returns the folder instance

        Arguments:
            folder_id {int} -- The ID of the folder to be retrieved

        Returns:
            Folder -- The folder
        """
        return Folder.get(folder_id, http_client=self.http_client)

    def get_or_create_folder(self, folder_path: Path, base_folder: Folder = None):
        """Creates a path in the base folder (base_folder_id).

        Arguments:
            folder_path {Path} -- A path to be created.

        Keyword Arguments:
            base_folder {Folder} -- The of the base folder. If None the root_folder is assumed. (default: {None})

        Returns:
            Folder -- Returns the last created folder
        """
        if base_folder is None:
            base_folder = self.get_root_folder()

        return base_folder.get_or_create(folder_path)

    # Patient
    def get_patients(self, filters=None):
        return Patient.get_list(filters, http_client=self.http_client)

    def get_patient(self, patient_id):
        return Patient.get(patient_id, http_client=self.http_client)

    # Exam
    def get_exam_list(self, filters=None):
        return Exam.get_list(filters, http_client=self.http_client)

    def get_exam(self, exam_id: int):
        return Exam.get(exam_id, http_client=self.http_client)

    # Series
    def get_series_list(self, filters=None):
        return Series.get_list(filters, http_client=self.http_client)

    def get_series(self, series_id):
        return Series.get(series_id, http_client=self.http_client)

    def get_dataset(self, dataset_id):
        return Dataset.get(dataset_id, http_client=self.http_client)

    def get_task(self, task_id):
        return Task.get(task_id, http_client=self.http_client)

    def delete_task(self, task_id):
        url = f'{Task.BASE_URL}{task_id}/'
        response = self.http_client.delete(url, timeout=60)
        if response.status_code != 204:
            raise AgoraException('Cannot delete the task: ' + response.text)

    # Search
    def search_series(self, aSearchString):

        return Series.search(aSearchString, self.http_client)

    # Import
    def upload(self, paths: List[Path], target_folder_id: Folder = None, json_import_file: Path = None, wait=True,
               progress=False, relations: dict =None):
        """Upload and import files to Agora

        Arguments:
            paths {List[Path]} -- A list of files or directories to upload (default: {None})

        Keyword Arguments:
            target_folder {Folder} -- The destination folder (default: {None})
            json_import_file {Path} -- The path to a JSON import file. Will be used only in special cases.
                                       (default: {None})
            wait {bool} -- Wait until the full upload has been finished (default: {True})
            progress {bool} -- Show a progress (default: {False})

        Returns:
            [type] -- [description]
        """
        if relations:
            self.version.needs('6.3.0', 'relations')

        for path in paths:
            if not path.exists():
                raise FileNotFoundError(path.as_posix())

        return import_data(self.http_client, paths=paths, target_folder_id=target_folder_id,
                           json_import_file=json_import_file, wait=wait, progress=progress, relations=relations)

    def import_directory(self, directory: Path, target_folder_id: int = None, json_import_file: Path = None, wait=True,
                         progress=False, relations: dict =None):
        """Upload and import a directory to Agora.

        The directory sturcture of all subdirectries will be preserved.

        Arguments:
            directory {Path} -- A single directory to upload

        Keyword Arguments:
            target_folder {Folder} -- [description] (default: {None})
            json_import_file {Path} -- [description] (default: {None})
            wait {bool} -- [description] (default: {True})
            progress {bool} -- [description] (default: {False})
        """
        if directory is None or not isinstance(directory, Path) or directory.is_dir is False:
            raise TypeError("Expects a pathlib.Path for directory")

        if directory.exists is False:
            raise FileNotFoundError("Directory doesn't exists")

        return import_data(self.http_client, paths=[directory], target_folder_id=target_folder_id, json_import_file=json_import_file, wait=wait, progress=progress, relations=relations)

    # User
    def get_users(self):
        """Returns a list of all Agora users

        Returns:
            List[User] -- The list of users
        """
        return User.get_list(http_client=self.http_client)

    def get_current_user(self):
        """Returns the current user

        Returns:
            User -- The user object of the current user
        """
        return User.get_current_user(http_client=self.http_client)

    def get_groups(self):
        """Returns a list of all gropus

        Returns:
            List[Group] -- A list of a groups
        """
        return Group.get_list(http_client=self.http_client)

    def get_api_key(self, create=False):
        """Returns the API Key of the current user

        Keyword Arguments:
            create {bool} -- Creates a new API key even if there is a already an API key (default: {False})

        Raises:
            AgoraException: [description]

        Returns:
            [type] -- [description]
        """
        response = self.http_client.get('/api/v1/apikey/')

        if response.status_code == 404 and create:
            response = self.http_client.post('/api/v1/apikey/', json={})

        if response.status_code == 200 or response.status_code == 201:
            data = response.json()
            return data

        raise AgoraException('Could not get the API key')

    def get_version(self):
        """Returns the Agora version

        Returns:
            Version -- A Version object
        """
        return Version.get(http_client=self.http_client)

    def get_version(self):
        """Returns the Agora version

        Returns:
            Version -- A Version object
        """
        return Version.get(http_client=self.http_client)

    def get_tags(self):
        return Tag.get_list(http_client=self.http_client)

    def get_tag(self, id=None, name=None):
        if not id and not name:
            raise AgoraException('please give a name or id as input')

        if id:
            return Tag.get(id, http_client=self.http_client)
        elif name:
            tags = self.get_tags()
            return next((x for x in tags if x.label == name), None)


    def close(self):
        pass



import json

from gtagora.exception import AgoraException
from gtagora.models.exam import Exam
from gtagora.models.host import Host
from gtagora.models.base import BaseModel
from gtagora.models.dataset import Dataset
from gtagora.models.folder import Folder
from gtagora.models.project_role import ProjectRole
from gtagora.models.series import Series
from gtagora.models.task import Task
from gtagora.models.trash import Trash
from gtagora.utils import remove_illegal_chars

from pathlib import Path


class Project(BaseModel):
    BASE_URL = '/api/v2/project/'

    ROLE_MANAGER = 1
    ROLE_SCIENTIST = 2
    ROLE_USER = 3
    ROLE_OBSERVER = 4

    PROJECT_ROLE = (
        (ROLE_MANAGER, 'manager'),
        (ROLE_SCIENTIST, 'scientist'),
        (ROLE_USER, 'user'),
        (ROLE_OBSERVER, 'observer')
        )

    def set_name(self, name):
        url = self.BASE_URL + str(self.id) + '/'
        data = {"name": name}
        response = self.http_client.put(url, data)

        if response.status_code == 200:
            data = response.json()
            self._set_values(data)
            return self
        else:
            raise AgoraException('Could not set the project name {0}', response.status_code)

    def get_exams(self, filters=None):
        if filters and not isinstance(filters, dict):
            raise AgoraException('The filter must be a dict')

        url = f'{self.BASE_URL}{self.id}/exam/?limit=10000000000'
        return self._get_object_list(url, filters, Exam)

    def get_tasks(self):
        url = f'{self.BASE_URL}{self.id}/task/?limit=10000000000'
        return self._get_object_list(url, None, Task)

    def import_tasks(self, file):
        with open(file) as json_file:
            data = json.load(json_file)

        url = f'{self.BASE_URL}{self.id}/task/imp/'
        response = self.http_client.post(url, json=data, timeout=60)
        if response.status_code != 200:
            raise AgoraException(f'Could not import the task: status = {response.status_code}')

    def get_hosts(self):
        url = f'{self.BASE_URL}{self.id}/host/?limit=10000000000'
        return self._get_object_list(url, None, Host)

    def get_root_folder(self):
        return Folder.get(self.root_folder, http_client=self.http_client)

    def add_member(self, user_id: int, role: int):
        role_id = None
        roles = self.get_roles()
        for r in roles:
            if r.id == role:
                role_id = r.id
                break
        if not role_id:
            raise AgoraException('Unknown role')

        url = f'/api/v2/projectmembership/'
        data = dict()
        data['project'] = self.id
        data['role'] = role_id
        data['user'] = user_id
        response = self.http_client.post(url, json=data, timeout=60)
        if response.status_code != 201:
            raise AgoraException(f'Could not add member: status = {response.status_code}')

    def copy_settings_from(self, other_project, copy_members=True, copy_tasks=True, copy_hosts=True):
        if copy_members:
            for membership in other_project.memberships:
                self.add_member(membership.get('user'), membership.get('role'))

        if copy_tasks:
            tasks = other_project.get_tasks()
            for task in tasks:
                task.copy_to_project(self.id)

        if copy_hosts:
            hosts = other_project.get_hosts()
            for host in hosts:
                host.copy_to_project(self.id)

    def get_roles(self):
        return self._get_object_list(ProjectRole.BASE_URL, None, ProjectRole)

    # Trash
    def empty_trash(self):
        trash = Trash()
        trash.empty(self.id)

    @property
    def display_name(self):
        if self.is_myagora:
            return 'My Agora'
        else:
            return self.name

    def __str__(self):
        return f"Project: {self.get_display_name()}"



import requests


class ApiKeyAuth(requests.auth.AuthBase):

    def __init__(self, api_key=None):
        self.api_key = api_key

    def __call__(self, r):
        r.headers['Authorization'] = 'X-Agora-Api-Key ' + self.api_key
        return r


class TokenAuth(requests.auth.AuthBase):

    def __init__(self, token=None):
        self.token = token

    def __call__(self, r):
        r.headers['Authorization'] = 'Token ' + self.token
        return r


class NoAuth(requests.auth.AuthBase):

    def __call__(self, r):
        return r


from gtagora.exception import AgoraException
from gtagora.models.base import BaseModel, DownloadDatasetMixin, TagMixin
from gtagora.models.exam import Exam


class Patient(BaseModel, TagMixin, DownloadDatasetMixin):

    BASE_URL = '/api/v1/patient/'

    def get_exams(self, filters=None):
        if filters and not isinstance(filters, dict):
            raise AgoraException('The filter must be a dict')

        url = f'{self.BASE_URL}{self.id}/exams/?limit=10000000000'
        return self._get_object_list(url, filters, Exam)

    def get_series(self, filters=None):
        if filters and not isinstance(filters, dict):
            raise AgoraException('The filter must be a dict')

        series = []
        exams = self.get_exams()
        for exam in exams:
            series += exam.get_series(filters)

        return series

    def get_datasets(self, filters=None):
        if filters and not isinstance(filters, dict):
            raise AgoraException('The filter must be a dict')

        datasets = []
        series = self.get_series()
        for s in series:
            datasets += s.get_datasets(filters)

        return datasets



from gtagora.exception import AgoraException
from gtagora.models.base import BaseModel
from gtagora.http.client import Client
from gtagora.models.base import get_client


class User(BaseModel):
    BASE_URL = '/api/v1/user/'

    @classmethod
    def get_current_user(cls, http_client=None):
        url = f'{cls.BASE_URL}current'
        http_client = get_client(http_client)

        response = http_client.get(url)

        if response.status_code == 200:
            data = response.json()
            return cls.from_response(data, http_client)

        raise AgoraException('Could not get the current user')

    @classmethod
    def get_or_create(cls, username, password, email=None, first_name=None, last_name=None, http_client=None):
        http_client = get_client(http_client)
        url = cls.BASE_URL

        data = cls.get_list({'username': username})
        for user in data:
            if user.username == username:
                return user, True

        data = {
            'username': username,
            'password': password,
            'email': email,
            'first_name': first_name,
            'last_name': last_name
        }

        response = http_client.post(url, data, timeout=60)
        if response.status_code == 201:
            data = response.json()
            if 'id' in data:
                new_user = User.from_response(data, http_client)
                return new_user, False

        raise AgoraException('Could not create the user')

    def delete(self):
        raise AgoraException("Can't delete user via the python interface")

    def __str__(self):
        return f"User {self.id}: {self.username}"



import math
import os
import uuid

import requests

from gtagora.exception import AgoraException


class Client:
    TIMEOUT = 20
    UPLOAD_TIMEOUT = 60
    UPLOAD_CHUCK_SIZE = 100 * 1024 * 1024  # 100MB
    DOWNLOAD_CHUNK_SIZE = 1024 * 1024  # 1MB

    def __init__(self, connection):
        self.connection = connection

    def check_connection(self):
        response = self.get('/api/v1/user/current/')
        if response.status_code == 200:
            data = response.json()
            return 'institution' in data

        return False

    def get(self, url, timeout=None, params=None, **kwargs):
        url = self.connection.url + url
        timeout = timeout if timeout else self.TIMEOUT
        auth = self.connection.get_auth()
        return requests.get(url, auth=auth, params=params, timeout=timeout,
                            verify=self.connection.verify_certificate, **kwargs)

    def post(self, url, data=None, json=None, timeout=None, params=None, **kwargs):
        url = self.connection.url + url
        timeout = timeout if timeout else self.TIMEOUT
        return requests.post(url, auth=self.connection.get_auth(), data=data, json=json, params=params, timeout=timeout,
                             verify=self.connection.verify_certificate, **kwargs)

    def put(self, url, json, timeout=None, params=None, **kwargs):
        url = self.connection.url + url
        timeout = timeout if timeout else self.TIMEOUT
        return requests.put(url, auth=self.connection.get_auth(), json=json, params=params, timeout=timeout,
                            verify=self.connection.verify_certificate, **kwargs)

    def delete(self, url, timeout=None, **kwargs):
        url = self.connection.url + url
        timeout = timeout if timeout else self.TIMEOUT
        return requests.delete(url, auth=self.connection.get_auth(), timeout=timeout,
                               verify=self.connection.verify_certificate, **kwargs)

    def download(self, url, target_filename):
        response = self.get(url, stream=True)
        if response.status_code == 200:
            with open(target_filename, 'wb') as file:
                for chunk in response.iter_content(self.DOWNLOAD_CHUNK_SIZE):
                    file.write(chunk)

    def upload(self, url, input_files, target_files=None, progress=False):
        response = self.get('/api/v1/version/')
        if response.status_code == 200:
            data = response.json()
            print(f"Ping {data}")
        else:
            print("Ping failed")

        if isinstance(input_files, str):
            files = []
            files.append(input_files)
        else:
            files = input_files

        if not target_files:
            target_files = [os.path.basename(file) for file in files]

        if len(target_files) < len(files):
            raise AgoraException('target_files list too short.')

        for index, cur_file in enumerate(files):
            print(f"Upload file: {cur_file} > {url}")

            if not os.path.isfile(cur_file):
                raise AgoraException('Could not open file ' + cur_file)

            filesize = os.path.getsize(cur_file)
            nof_chunks = math.ceil(filesize / self.UPLOAD_CHUCK_SIZE)

            head, tail = os.path.split(cur_file)
            filename = tail
            target_filename = target_files[index]

            uid = str(uuid.uuid4())

            with open(cur_file, mode='rb') as file:
                for chunk in range(0, nof_chunks):
                    if progress:
                        self.print_progress(index, len(files), chunk, nof_chunks)
                    data = file.read(self.UPLOAD_CHUCK_SIZE)
                    files = {'file': (filename, data)}
                    form = {
                        'description': '',
                        'flowChunkNumber': str(chunk),
                        'flowChunkSize': str(self.UPLOAD_CHUCK_SIZE),
                        'flowCurrentChunkSize': str(len(data)),
                        'flowTotalSize': str(filesize),
                        'flowIdentifier': uid,
                        'flowFilename': target_filename,
                        'flowRelativePath': target_filename,
                        'flowTotalChunks': str(nof_chunks)}
                    response = self.post(url, files=files, data=form, timeout=self.UPLOAD_TIMEOUT)
                    if response.status_code != 200:
                        raise AgoraException(
                            f"Failed to upload chunk {chunk} of file {cur_file}. Status code: {response.status_code}")

        print(f"Upload done")
        return True

    def print_progress(self, curFile, nof_file, chunk, nof_chunks):
        length = 40
        done = int((curFile + 1) / nof_file * length)
        bar = 'X' * done + '-' * (length - done)
        print('\r%s file %d of %d, chunk %d of %d' % (bar, curFile + 1, nof_file, chunk + 1, nof_chunks), end='\r')
        if curFile + 1 == nof_file and chunk + 1 == nof_chunks:
            print()



from gtagora.exception import AgoraException
from gtagora.models.base import BaseModel, LinkToFolderMixin, DownloadDatasetMixin, TagMixin
from gtagora.models.dataset import Dataset
from gtagora.models.series import Series
from gtagora.models.timeline import TimelineItem
from gtagora.utils import remove_illegal_chars

from pathlib import Path


class Exam(LinkToFolderMixin, DownloadDatasetMixin, TagMixin, BaseModel):
    BASE_URL = '/api/v1/exam/'
    BASE_URL_V2 = '/api/v2/exam/'

    def set_name(self, name):
        url = self.BASE_URL + str(self.id) + '/'
        data = {"name": name}
        response = self.http_client.put(url, data)

        if response.status_code == 200:
            data = response.json()
            self._set_values(data)
            return self
        else:
            raise AgoraException('Could not set the exam name {0}', response.status_code)

    def get_series(self, filters=None):
        if filters and not isinstance(filters, dict):
            raise AgoraException('The filter must be a dict')

        url = f'{self.BASE_URL}{self.id}/series/?limit=10000000000'
        return self._get_object_list(url, filters, Series)

    def get_datasets(self, filters=None):
        if filters and not isinstance(filters, dict):
            raise AgoraException('The filter must be a dict')

        series = self.get_series()
        datasets = []
        for s in series:
            datasets += s.get_datasets(filters)

        datasets += self.get_files()

        return datasets

    def get_files(self):
        files = []

        url = f'{self.BASE_URL}{self.id}/files/?limit=10000000000'
        response = self.http_client.get(url)
        if response.status_code == 200:
            data = response.json()

            if 'results' in data and 'count' in data:
                results = data['results']
                if data['count'] == 0:
                    return files
                if data['count'] != len(results):
                    print('Warning: Could not get all files')

                for curFile in results:
                    files.append(Dataset(curFile, http_client=self.http_client))

            return files

        raise AgoraException('Could not get the series')

    def upload(self, input_files, target_files=None):
        if target_files and len(input_files) != len(target_files):
            raise AgoraException("The Inputfiles and TargetFiles must have the same length")

        if isinstance(input_files, str):
            files = []
            files.append(input_files)
        else:
            files = input_files

        datasets = []
        for index, currrent_file in enumerate(files):
            cur_target_file = None
            if target_files:
                cur_target_file = target_files[index]
            datasets.append(self.http_client.upload_dataset(currrent_file, cur_target_file, self.http_client,
                                                            exam_id=self.id))
        return datasets

    def download(self, target_path: Path):
        for series in self.get_series():
            for dataset in series.get_datasets():
                final_path = target_path / remove_illegal_chars(self.name) / remove_illegal_chars(series.name)
                final_path.mkdir(parents=True, exist_ok=True)
                dataset.download(final_path)
        for dataset in self.get_files():
            final_path = target_path / self.name
            final_path.mkdir(parents=True, exist_ok=True)
            dataset.download(final_path)

    def upload_dataset(self, input_files, dataset_type, target_files=None):
        # This function creates a dataset of a given type all files given as input will be added to one dataset.
        # Please note: At the moment there is no consistency check. We could create datasets with improper
        # files (e.g. a PAR/REC dataset without PAR/REC files)
        return self.http_client.upload_dataset(input_files, target_files, self.http_client, exam_id=self.id,
                                               dataset_type=dataset_type)

    def copy_to_project(self, project_id):
        url = f'{self.BASE_URL_V2}{self.id}/copy_to/{project_id}/'
        response = self.http_client.post(url, json={}, timeout=60)
        if response.status_code != 200:
            raise AgoraException(f'Could not copy the exam: status = {response.status_code}')

        return self._get_new_exam_from_timeline(response)

    def copy_to_folder(self, target_folder_id):
        from gtagora.models.folder import Folder
        folder_url = f'{Folder.BASE_URL}{target_folder_id}/'
        response = self.http_client.get(folder_url, timeout=60)
        if response.status_code != 200:
            raise AgoraException(f'Could not find the folder with id = {target_folder_id}')
        folder = Folder.from_response(response.json(), self.http_client)

        url = f'{self.BASE_URL_V2}{self.id}/copy_to/{folder.project}/folder/{target_folder_id}/'

        response = self.http_client.post(url, json={}, timeout=60)
        if response.status_code != 200:
            raise AgoraException(f'Could not copy the exam: status = {response.status_code}')

        return self._get_new_exam_from_timeline(response)

    def move_to_project(self, project_id, target_folder_id=None):
        if not target_folder_id:
            url = f'{self.BASE_URL_V2}{self.id}/move_to{project_id}/'
        else:
            url = f'{self.BASE_URL_V2}{self.id}/move_to/{project_id}/folder/{target_folder_id}/'

        response = self.http_client.post(url, json={}, timeout=60)
        if response.status_code != 200:
            raise AgoraException(f'Could not move the exam: status = {response.status_code}')

        return self._get_new_exam_from_timeline(response)

    def __str__(self):
        return f"Exam: {self.name}"

    def _get_new_exam_from_timeline(self, response):
        timeline_item = TimelineItem.from_response(response.json(), self.http_client)
        timeline_item = timeline_item.poll()
        related_objects = timeline_item.data.get('related_objects')
        if related_objects:
            for obj in related_objects:
                if  obj.get('id') and obj.get('id') != self.id:
                    try:
                        return self.get(obj.get('id'), self.http_client)
                    except:
                        pass

        return None




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gtagora-connector",
    version="1.3",
    author="Martin Bührer, Felix Eichenberger",
    author_email="info@gyrotools.com",
    description="The Agora Connector for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gyrotools/gtagora",
    packages=setuptools.find_packages(),
    install_requires=[
        'requests>=2.0',
    ],
    python_requires='>=3.6.0',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



# from gtagora.models.folder import Folder
# from gtagora.models.folder_item import FolderItem
# from gtagora.models.exam import Exam
# from gtagora.models.series import Series
# from gtagora.models.dataset import Dataset
# from gtagora.models.datafile import Datafile
# from gtagora.models.parameter import Parameter
# from gtagora.models.patient import Patient



import json
from pathlib import Path

from gtagora.exception import AgoraException
from gtagora.models.base import BaseModel
from gtagora.models.timeline import TimelineItem


class Task(BaseModel):

    BASE_URL = '/api/v1/taskdefinition/'

    INPUT_TYPE_EXAM = 1
    INPUT_TYPE_SERIES = 2
    INPUT_TYPE_DATASET = 3
    INPUT_TYPE_FILE = 4
    INPUT_TYPE_STRING = 5
    INPUT_TYPE_INTEGER = 6
    INPUT_TYPE_FLOAT = 7
    INPUT_TYPE_SELECT = 8
    INPUT_TYPE_FOLDER = 9

    def run(self, input=None, target: BaseModel=None, **kwargs):
        if input:
            input_dict = self._get_inputs(input)
        else:
            input_dict = self._get_inputs(kwargs)
        self._check_outputs(target)

        data = {}

        if target:
            object_name = target.__class__.__name__.lower()
            data['target'] = {'object_id': target.id, 'object_type': object_name}
        else:
            data['target'] = None

        data['inputs'] = input_dict

        url = f'{self.BASE_URL}{self.id}/run/'
        response = self.http_client.post(url, json=data, timeout=60)

        if response.status_code != 200:
            raise AgoraException('Cannot run the task: ' + response.text)
        else:
            taskinfo = json.loads(response.content)
            result = TimelineItem.get_list_from_data([taskinfo])
            return result[0] if result else None

    def create(self):
        data = self.toDict()
        data['id'] = None
        if 'inputs' in data and data['inputs']:
            for input in data['inputs']:
                if 'id' in input:
                    input['id'] = None
        response = self.http_client.post(self.BASE_URL, json=data, timeout=60)
        if response.status_code != 201:
            raise AgoraException('Cannot create a task: ' + response.text)
        else:
            created_task = json.loads(response.content)
            result = self.get_list_from_data([created_task])
            return result[0] if result else None


    def save(self):
        if not hasattr(self, 'id') or not self.id:
            self.create()
        else:
            url = f'{self.BASE_URL}{self.id}/'
            data = self.toDict()
            response = self.http_client.put(url, json=data, timeout=60)
            if response.status_code != 200:
                raise AgoraException('Cannot create a task: ' + response.text)

    def delete(self):
        if not hasattr(self, 'id') or not self.id:
            raise AgoraException('Cannot delete the task: No ID available')
        else:
            url = f'{self.BASE_URL}{self.id}/'
            response = self.http_client.delete(url, timeout=60)
            if response.status_code != 204:
                raise AgoraException('Cannot delete the task: ' + response.text)

    def toDict(self):
        fields = ['container_name', 'container_options', 'execute_template', 'host', 'host_id', 'id', 'inputs', 'members', 'mount_volumes', 'name', 'outputs', 'parse_output_for_error', 'success_exit_code', 'task_target', 'use_docker']

        d = dict()
        for field in fields:
            d[field] = self.__dict__.get(field)

        return d

    def syntax(self):
        print(self._get_run_cmd())

    def export(self, dir: [str, Path]=None, file: [str, Path]=None):
        if not dir and not file:
            raise AgoraException(f'Please pass an output directory or an output filepath as argument')

        if dir and isinstance(dir, str):
            dir = Path(dir)

        if file and isinstance(file, str):
            file = Path(file)

        if dir and not dir.exists():
            raise AgoraException(f'The output directory does not exist')

        url = f'{self.BASE_URL}{self.id}/export/'
        response = self.http_client.get(url, timeout=60)
        if response.status_code != 200:
            raise AgoraException(f'Cannot export the task: status = {response.status_code}')
        data = response.json()

        if not file:
            filename = data.get('name')
            if not filename:
                raise AgoraException(f'Cannot determine the name of the task')
            filename = filename.replace("/", "_")
            filename = self.to_valid_filename(filename) + '.json'
            file = Path(dir) / Path(filename)

        with open(str(file), 'w') as fp:
            json.dump(data, fp)

    def copy_to_project(self, project_id):
        url = f'/api/v2/project/{project_id}/task/{self.id}/copy/'
        response = self.http_client.post(url, json={}, timeout=60)

        if response.status_code != 200:
            raise AgoraException(f'Cannot copy the task: status = {response.status_code}')

    def move_to_project(self, project_id):
        url = f'/api/v2/project/{project_id}/task/{self.id}/move/'
        response = self.http_client.post(url, json={}, timeout=60)

        if response.status_code != 200:
            raise AgoraException(f'Cannot copy the task: status = {response.status_code}')

    def _get_run_cmd(self):
        cmd = 'task.run('
        alternate_cmd = 'task.run('
        first = True
        for input in self.inputs:
            if not first:
                cmd += ', '
                alternate_cmd += ', '

            cmd += input['key'] + '=<' + self._get_type_name(input['type']) + '>'
            if input['type'] < 5:
                alternate_cmd += input['key'] + '=<' + self._get_type_name(input['type']) + ' ID>'
            else:
                alternate_cmd += input['key'] + '=<' + self._get_type_name(input['type']) + '>'

            first = False

        if self.outputs:
            if not first:
                cmd += ', '
                alternate_cmd += ', '
            cmd += 'target=<target>'
            alternate_cmd += 'target=<target>'
        cmd += ')'
        alternate_cmd += ')'
        return cmd + '\nOR\n' + alternate_cmd

    def _get_inputs(self, arguments):
        input_dict = {}
        for input in self.inputs:
            if not input['key'] in arguments:
                raise AgoraException('\n\nThe task input \'' + input['key'] + '\' is unassigned.\nRun the task with the following command:\n\n' + self._get_run_cmd())

            argument_name = input['key']
            argument = arguments[argument_name]
            argument_type = argument.__class__.__name__.lower()
            input_type = self._get_type_name(input['type'])
            if isinstance(argument, BaseModel):
                if argument_type != self._get_type_name(input['type']):
                    self._raise_input_error(input)
                input_dict[argument_name] = {'object_id': argument.id, 'object_type': argument_type}

            elif input['type'] < Task.INPUT_TYPE_STRING or input['type'] == Task.INPUT_TYPE_FOLDER:
                if not isinstance(argument, int):
                    self._raise_input_error(input)
                input_dict[argument_name] = {'object_id': argument, 'object_type': input_type}
            elif input['type'] == Task.INPUT_TYPE_STRING:
                if not isinstance(argument, str):
                    self._raise_input_error(input)
                input_dict[argument_name] = argument
            elif input['type'] == Task.INPUT_TYPE_INTEGER:
                if not isinstance(argument, int):
                    self._raise_input_error(input)
                input_dict[argument_name] = argument
            elif input['type'] == Task.INPUT_TYPE_FLOAT:
                if not isinstance(argument, float):
                    self._raise_input_error(input)
                input_dict[argument_name] = argument

        return input_dict

    def _check_outputs(self, target):
        if self.outputs and not target:
            raise AgoraException('\n\nThe \'target\' argument is missing (e.g. the output folder). Run the task with the following command:\n\n' + self._get_run_cmd())

        if self.outputs and not isinstance(target, BaseModel):
            raise AgoraException('\n\nThe target must be an Agora object (e.g. a folder)')

    def _raise_input_error(self, input):
        argument_name = input['key']
        argument_type = self._get_type_name(input['type'])
        msg = ''
        if input['type'] < Task.INPUT_TYPE_STRING or input['type'] == Task.INPUT_TYPE_FOLDER:
            msg += '\n\nThe task input \'' + argument_name + '\' must eihter be a ' + argument_type + ' or a ' + argument_type + ' ID'
        else:
            msg += '\n\nThe task input \'' + argument_name + '\' must be a ' + argument_type

        msg += '\n\nRun the task with the following syntax:\n' + self._get_run_cmd()

        raise AgoraException(msg)



    @staticmethod
    def _get_type_name(type: int):
        if type == 0:
            return 'none'
        elif type == Task.INPUT_TYPE_EXAM:
            return 'exam'
        elif type == Task.INPUT_TYPE_SERIES:
            return 'series'
        elif type == Task.INPUT_TYPE_DATASET:
            return 'dataset'
        elif type == Task.INPUT_TYPE_FOLDER:
            return 'folder'
        elif type == Task.INPUT_TYPE_STRING:
            return 'string'
        elif type == Task.INPUT_TYPE_INTEGER:
            return 'integer'
        elif type == Task.INPUT_TYPE_FLOAT:
            return 'float'
        elif type == Task.INPUT_TYPE_SELECT:
            return 'select'




import pprint
import unicodedata
import re

from gtagora.exception import AgoraException


def get_client(http_client):
    from gtagora import Agora

    return http_client if http_client else Agora.default_client


class BaseModel:

    BASE_URL = ''

    def __init__(self, http_client=None):
        self.http_client = get_client(http_client)

    @classmethod
    def from_response(cls, model_dict, http_client=None):
        instance = cls(http_client=http_client)
        instance._set_values(model_dict)

        return instance

    @classmethod
    def get(cls, id=None, http_client=None):
        instance = cls(http_client=http_client)
        return instance._get_object(id)

    @classmethod
    def get_list(cls, filters=None, http_client=None):
        if filters and not isinstance(filters, dict):
            raise AgoraException('The filter must be a dict')

        instance = cls(http_client=http_client)
        filters = filters if filters else {}
        if 'limit' not in filters:
            filters['limit'] = '10000000000'

        url = cls.BASE_URL
        return instance._get_object_list(url, filters, cls)

    @classmethod
    def get_list_from_data(cls, data):
        object_list = []

        if 'results' in data and 'count' in data:
            results = data['results']
            if data['count'] == 0:
                return object_list
            if data['count'] != len(results):
                print('Warning: Could not get all series')

            for r in results:
                object_list.append(cls.from_response(r))
        elif isinstance(data, list):
            object_list = [cls.from_response(d) for d in data]

        return object_list

        raise AgoraException(f'Could not get the {cls.__name__} list')

    def delete(self):
        url = f'{self.BASE_URL}{self.id}/'
        response = self.http_client.delete(url)

        if response.status_code == 204:
            return True
        raise AgoraException('Could not delete FolderItem')

    def to_valid_filename(self, value, allow_unicode=False):
        """
        Taken from https://github.com/django/django/blob/master/django/utils/text.py
        Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
        dashes to single dashes. Remove characters that aren't alphanumerics,
        underscores, or hyphens. Convert to lowercase. Also strip leading and
        trailing whitespace, dashes, and underscores.
        """
        value = str(value)
        if allow_unicode:
            value = unicodedata.normalize('NFKC', value)
        else:
            value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
        value = re.sub(r'[^\w\s-]', '', value.lower())
        return re.sub(r'[-\s]+', '-', value).strip('-_')

    def to_dict(self):
        excluded_keys = ['http_client']
        return dict( (key, value) for (key, value) in self.__dict__.items() if key not in excluded_keys and not key.startswith('__'))

    def content_type(self):
        from gtagora.models.exam import Exam
        from gtagora.models.folder import Folder
        from gtagora.models.series import Series
        from gtagora.models.dataset import Dataset
        from gtagora.models.patient import Patient

        if isinstance(self, Exam):
            return 'exam'
        elif isinstance(self, Folder):
            return 'folder'
        elif isinstance(self, Series):
            return 'serie'
        elif isinstance(self, Dataset):
            return 'dataset'
        elif isinstance(self, Patient):
            return 'patient'

    def _set_values(self, model_dict):
        for key, value in model_dict.items():
            setattr(self, key, value)

    def _get_object(self, id):
        if id:
            url = f'{self.BASE_URL}{id}/'
        else:
            url = f'{self.BASE_URL}'

        response = self.http_client.get(url)
        if response.status_code == 200:
            data = response.json()
            return self.__class__.from_response(data, http_client=self.http_client)

        raise AgoraException('Could not get the {0}. HTTP status = {1}'.format(self.__class__.__name__, response.status_code))

    def _get_object_list(self, url, params, object_class):
        response = self.http_client.get(url, params=params)

        if response.status_code == 200:
            data = response.json()
            object_list = []

            if 'results' in data and 'count' in data:
                results = data['results']
                if data['count'] == 0:
                    return object_list
                if data['count'] != len(results):
                    print('Warning: Could not get all series')

                for r in results:
                    object_list.append(object_class.from_response(r))
            elif isinstance(data, list):
                object_list = [object_class.from_response(d) for d in data]
            else:
                return object_class.from_response(data)

            return object_list

        raise AgoraException(f'Could not get the {object_class.__name__} list')

    def display(self):
        pprint.pprint(vars(self), indent=4)


class DownloadDatasetMixin:

    def download(self, filename):
        datasets = self.get_datasets()
        return [dataset.download(filename) for dataset in datasets]


class LinkToFolderMixin:
    def link_to_folder(self, folder):
        from gtagora.models.folder_item import FolderItem
        from gtagora.models.folder import Folder

        if isinstance(folder, Folder):
            folder_id = folder.id
        elif isinstance(folder, int):
            folder_id = folder
        else:
            raise AgoraException('The input must either be a folder or a folder id')

        url = f'{self.BASE_URL}{self.id}/link_to/{folder_id}/'
        post_data = {}
        response = self.http_client.post(url, post_data)
        if response.status_code == 201:
            return FolderItem.from_response(response.json(), http_client=self.http_client)

        raise AgoraException('Could not create a link')

    def get_folders(self, parent_folder = None, filters=None):
        from gtagora.models.folder import Folder

        if filters and not isinstance(filters, dict):
            raise AgoraException('The filter must be a dict')

        url = f'{self.BASE_URL}{self.id}/folders/?limit=10000000000'

        folders = self._get_object_list(url, filters, Folder)

        if parent_folder:
            filtered_folders = []
            for f in folders:
                if f.is_subfolder_of(parent_folder):
                    filtered_folders.append(f)

            folders = filtered_folders if filtered_folders else None

        return folders

    def is_in_folder(self, object: BaseModel, folder):
        from gtagora.models.folder import Folder

        folder_id = None
        if isinstance(folder, Folder):
            folder_id = folder.id
        elif isinstance(folder, int):
            folder_id = folder
        else:
            raise AgoraException('The folder argument must either be a Folder class or a folder ID')

        folders = object.get_folders()
        for f in folders:
            breadcrumb = f.get_breadcrumb()
            for b in breadcrumb:
                if b.object_id == folder_id:
                    return f

        return None


class SearchMixin:

    @classmethod
    def search(cls, search_string, http_client=None):
        instance = cls(http_client=http_client)

        if not isinstance(search_string, str):
            raise AgoraException('The search term must be a string')

        url = f'{instance.BASE_URL}search/?q=' + search_string + '&limit=10000000000'
        response = http_client.get(url)
        if response.status_code == 200:
            list = instance.get_list_from_data(response.json())
            return list
        else:
            raise AgoraException(f'Search unsuccessful: {response.text}')

        return []


class TagMixin:
    def tag(self, tag):
        from gtagora.models.tag import Tag
        from gtagora.models.tag import TagInstance

        if isinstance(tag, Tag):
            tag_id = tag.id
        elif isinstance(tag, int):
            tag_id = tag
        else:
            raise AgoraException('The input must either be a tag or a tag id')

        url = TagInstance.BASE_URL
        post_data = {'tag_definition': tag_id, 'tagged_object_content_type': self.content_type(), 'tagged_object_id': self.id}
        response = self.http_client.post(url, post_data)
        if response.status_code == 201:
            instances = TagInstance.get_list_from_data(response.json())
            if instances:
                return instances[0]
            else:
                return None

        raise AgoraException('Could not tag the object')



from gtagora.models.base import BaseModel
from distutils.version import LooseVersion
from gtagora.exception import AgoraException


class Version(BaseModel):

    BASE_URL = '/api/v1/version/'

    def is_dev(self):
        return self.version.lower() == 'dev'

    def is_higher_than(self, version):
        return LooseVersion(self.version) > LooseVersion(version)

    def is_lower_than(self, version):
        return LooseVersion(self.version) < LooseVersion(version)

    def needs(self, version, feature=None, error_message=None):
        if not self.is_dev() and self.is_lower_than(version):
            if error_message:
                msg = error_message
            elif feature:
                msg = f'The {feature} feature needs Agora version {version} or higher. Please upgrade your Agora'
            else:
                msg = f'This feature needs Agora version {version} or higher. Please upgrade your Agora'
            raise AgoraException(msg)




